@@ -95,6 +95,33 @@ static void luv_push_stats_table(lua_State* L, const uv_stat_t* s) {
   }
 }
 
+static int luv_push_dirent(lua_State* L, const uv_dirent_t* ent, int table) {
+  const char* type;
+  if (table) {
+    lua_newtable(L);
+  }
+  lua_pushstring(L, ent->name);
+  if (table) {
+    lua_setfield(L, -2, "name");
+  }
+  switch (ent->type) {
+    case UV_DIRENT_UNKNOWN: return 1;
+    case UV_DIRENT_FILE:    type = "file"; break;
+    case UV_DIRENT_DIR:     type = "directory"; break;
+    case UV_DIRENT_LINK:    type = "link"; break;
+    case UV_DIRENT_FIFO:    type = "fifo"; break;
+    case UV_DIRENT_SOCKET:  type = "socket"; break;
+    case UV_DIRENT_CHAR:    type = "char"; break;
+    case UV_DIRENT_BLOCK:   type = "block"; break;
+    default:                type = "unknown"; break;
+  }
+  lua_pushstring(L, type);
+  if (table)
+    lua_setfield(L, -2, "type");
+
+  return table ? 1 : 2;
+}
+
 static int luv_check_flags(lua_State* L, int index) {
   const char* string;
   if (lua_isnumber(L, index)) {
@@ -161,6 +188,26 @@ static int luv_check_amode(lua_State* L, int index) {
   return mode;
 }
 
+#if LUV_UV_VERSION_GEQ(1, 31, 0)
+static void luv_push_statfs_table(lua_State* L, const uv_statfs_t* s) {
+  lua_createtable(L, 0, 8);
+  lua_pushinteger(L, s->f_type);
+  lua_setfield(L, -2, "type");
+  lua_pushinteger(L, s->f_bsize);
+  lua_setfield(L, -2, "bsize");
+  lua_pushinteger(L, s->f_blocks);
+  lua_setfield(L, -2, "blocks");
+  lua_pushinteger(L, s->f_bfree);
+  lua_setfield(L, -2, "bfree");
+  lua_pushinteger(L, s->f_bavail);
+  lua_setfield(L, -2, "bavail");
+  lua_pushinteger(L, s->f_files);
+  lua_setfield(L, -2, "files");
+  lua_pushinteger(L, s->f_ffree);
+  lua_setfield(L, -2, "ffree");
+};
+#endif
+
 /* Processes a result and pushes the data onto the stack
    returns the number of items pushed */
 static int push_fs_result(lua_State* L, uv_fs_t* req) {
@@ -197,6 +244,9 @@ static int push_fs_result(lua_State* L, uv_fs_t* req) {
     case UV_FS_FCHMOD:
     case UV_FS_CHOWN:
     case UV_FS_FCHOWN:
+#if LUV_UV_VERSION_GEQ(1, 21, 0)
+    case UV_FS_LCHOWN:
+#endif
     case UV_FS_UTIME:
     case UV_FS_FUTIME:
     case UV_FS_COPYFILE:
@@ -215,6 +265,12 @@ static int push_fs_result(lua_State* L, uv_fs_t* req) {
       luv_push_stats_table(L, &req->statbuf);
       return 1;
 
+#if LUV_UV_VERSION_GEQ(1, 31, 0)
+    case UV_FS_STATFS:
+      luv_push_statfs_table(L, req->ptr);
+      return 1;
+#endif
+
     case UV_FS_MKDTEMP:
       lua_pushstring(L, req->path);
       return 1;
@@ -233,6 +289,46 @@ static int push_fs_result(lua_State* L, uv_fs_t* req) {
       lua_rawgeti(L, LUA_REGISTRYINDEX, data->req_ref);
       return 1;
 
+#if LUV_UV_VERSION_GEQ(1, 28, 0)
+    case UV_FS_OPENDIR: {
+      int nentries;
+      uv_dir_t* dir = (uv_dir_t*)req->ptr;
+
+      lua_rawgeti(L, LUA_REGISTRYINDEX, data->data_ref);
+      nentries = luaL_checkinteger(L, -1);
+      lua_pop(L, 1);
+      luaL_unref(L, LUA_REGISTRYINDEX, data->data_ref);
+      data->data_ref = LUA_NOREF;
+
+      (*(void**)lua_newuserdata(L, sizeof(void*))) = dir;
+      lua_pushfstring(L, "uv_dir:%p", dir);
+      dir->dirents = lua_newuserdata(L, sizeof(uv_dirent_t)*nentries);
+      dir->nentries = nentries;
+      lua_rawset(L, LUA_REGISTRYINDEX);
+      luaL_getmetatable(L, "uv_dir");
+      lua_setmetatable(L, -2);
+
+      return 1;
+    }
+    case UV_FS_READDIR: {
+      if(req->result > 0) {
+        size_t i;
+        uv_dir_t *dir = (uv_dir_t*)req->ptr;
+        lua_newtable(L);
+        for(i=0; i<req->result; i++) {
+          luv_push_dirent(L, dir->dirents+i, 1);
+          lua_rawseti(L, -2, i+1);
+        }
+      } else
+        lua_pushnil(L);
+
+      return 1;
+    }
+    case UV_FS_CLOSEDIR:
+      lua_pushboolean(L, 1);
+      return 1;
+#endif
+
     default:
       lua_pushnil(L);
       lua_pushfstring(L, "UNKNOWN FS TYPE %d\n", req->fs_type);
@@ -242,7 +338,8 @@ static int push_fs_result(lua_State* L, uv_fs_t* req) {
 }
 
 static void luv_fs_cb(uv_fs_t* req) {
-  lua_State* L = luv_state(req->loop);
+  luv_req_t* data = (luv_req_t*)req->data;
+  lua_State* L = data->ctx->L;
 
   int nargs = push_fs_result(L, req);
   if (nargs == 2 && lua_isnil(L, -nargs)) {
@@ -256,38 +353,48 @@ static void luv_fs_cb(uv_fs_t* req) {
     lua_insert(L, -nargs - 1);
     nargs++;
   }
-  luv_fulfill_req(L, (luv_req_t*)req->data, nargs);
-  if (req->fs_type != UV_FS_SCANDIR) {
-    luv_cleanup_req(L, (luv_req_t*)req->data);
-    req->data = NULL;
+  if (req->fs_type == UV_FS_SCANDIR) {
+    luv_fulfill_req(L, data, nargs);
+  }
+  else {
+    // cleanup the uv_fs_t before the callback is called to avoid
+    // a race condition when fs_close is called from within
+    // a fs_readdir callback, see https://github.com/luvit/luv/issues/384
     uv_fs_req_cleanup(req);
+    req->data = NULL;
+    luv_fulfill_req(L, data, nargs);
+    luv_cleanup_req(L, data);
   }
 }
 
 #define FS_CALL(func, req, ...) {                         \
   int ret, sync;                                          \
-  luv_req_t* data = (luv_req_t*)req->data;                            \
+  luv_req_t* data = (luv_req_t*)req->data;                \
   sync = data->callback_ref == LUA_NOREF;                 \
-  ret = uv_fs_##func(luv_loop(L), req, __VA_ARGS__,       \
+  ret = uv_fs_##func(data->ctx->loop, req, __VA_ARGS__,   \
                      sync ? NULL : luv_fs_cb);            \
   if (req->fs_type != UV_FS_ACCESS && ret < 0) {          \
     lua_pushnil(L);                                       \
     if (req->path) {                                      \
-      lua_pushfstring(L, "%s: %s: %s", uv_err_name(req->result), uv_strerror(req->result), req->path); \
+      lua_pushfstring(L, "%s: %s: %s",                    \
+          uv_err_name(req->result),                       \
+          uv_strerror(req->result), req->path);           \
     }                                                     \
     else {                                                \
-      lua_pushfstring(L, "%s: %s", uv_err_name(req->result), uv_strerror(req->result)); \
+      lua_pushfstring(L, "%s: %s",                        \
+          uv_err_name(req->result),                       \
+          uv_strerror(req->result));                      \
     }                                                     \
     lua_pushstring(L, uv_err_name(req->result));          \
-    luv_cleanup_req(L, (luv_req_t*)req->data);                        \
+    luv_cleanup_req(L, data);                             \
     req->data = NULL;                                     \
     uv_fs_req_cleanup(req);                               \
     return 3;                                             \
   }                                                       \
   if (sync) {                                             \
     int nargs = push_fs_result(L, req);                   \
-    if (req->fs_type != UV_FS_SCANDIR) {                  \
-      luv_cleanup_req(L, (luv_req_t*)req->data);                      \
+    if(req->fs_type != UV_FS_SCANDIR) {                   \
+      luv_cleanup_req(L, data);                           \
       req->data = NULL;                                   \
       uv_fs_req_cleanup(req);                             \
     }                                                     \
@@ -298,24 +405,27 @@ static void luv_fs_cb(uv_fs_t* req) {
 }
 
 static int luv_fs_close(lua_State* L) {
+  luv_ctx_t* ctx = luv_context(L);
   uv_file file = luaL_checkinteger(L, 1);
   int ref = luv_check_continuation(L, 2);
   uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, sizeof(*req));
-  req->data = luv_setup_req(L, ref);
+  req->data = luv_setup_req(L, ctx, ref);
   FS_CALL(close, req, file);
 }
 
 static int luv_fs_open(lua_State* L) {
+  luv_ctx_t* ctx = luv_context(L);
   const char* path = luaL_checkstring(L, 1);
   int flags = luv_check_flags(L, 2);
   int mode = luaL_checkinteger(L, 3);
   int ref = luv_check_continuation(L, 4);
   uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, sizeof(*req));
-  req->data = luv_setup_req(L, ref);
+  req->data = luv_setup_req(L, ctx, ref);
   FS_CALL(open, req, path, flags, mode);
 }
 
 static int luv_fs_read(lua_State* L) {
+  luv_ctx_t* ctx = luv_context(L);
   uv_file file = luaL_checkinteger(L, 1);
   int64_t len = luaL_checkinteger(L, 2);
   int64_t offset = luaL_checkinteger(L, 3);
@@ -327,21 +437,23 @@ static int luv_fs_read(lua_State* L) {
   buf = uv_buf_init(data, len);
   ref = luv_check_continuation(L, 4);
   req = (uv_fs_t*)lua_newuserdata(L, sizeof(*req));
-  req->data = luv_setup_req(L, ref);
+  req->data = luv_setup_req(L, ctx, ref);
   // TODO: find out why we can't just use req->ptr for the base
   ((luv_req_t*)req->data)->data = buf.base;
   FS_CALL(read, req, file, &buf, 1, offset);
 }
 
 static int luv_fs_unlink(lua_State* L) {
+  luv_ctx_t* ctx = luv_context(L);
   const char* path = luaL_checkstring(L, 1);
   int ref = luv_check_continuation(L, 2);
   uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, sizeof(*req));
-  req->data = luv_setup_req(L, ref);
+  req->data = luv_setup_req(L, ctx, ref);
   FS_CALL(unlink, req, path);
 }
 
 static int luv_fs_write(lua_State* L) {
+  luv_ctx_t* ctx = luv_context(L);
   uv_file file = luaL_checkinteger(L, 1);
   uv_buf_t buf;
   int64_t offset;
@@ -365,43 +477,47 @@ static int luv_fs_write(lua_State* L) {
   offset = luaL_checkinteger(L, 3);
   ref = luv_check_continuation(L, 4);
   req = (uv_fs_t*)lua_newuserdata(L, sizeof(*req));
-  req->data = luv_setup_req(L, ref);
+  req->data = luv_setup_req(L, ctx, ref);
   req->ptr = buf.base;
   ((luv_req_t*)req->data)->data = bufs;
   FS_CALL(write, req, file, bufs ? bufs : &buf, count, offset);
 }
 
 static int luv_fs_mkdir(lua_State* L) {
+  luv_ctx_t* ctx = luv_context(L);
   const char* path = luaL_checkstring(L, 1);
   int mode = luaL_checkinteger(L, 2);
   int ref = luv_check_continuation(L, 3);
   uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, sizeof(*req));
-  req->data = luv_setup_req(L, ref);
+  req->data = luv_setup_req(L, ctx, ref);
   FS_CALL(mkdir, req, path, mode);
 }
 
 static int luv_fs_mkdtemp(lua_State* L) {
+  luv_ctx_t* ctx = luv_context(L);
   const char* tpl = luaL_checkstring(L, 1);
   int ref = luv_check_continuation(L, 2);
   uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, sizeof(*req));
-  req->data = luv_setup_req(L, ref);
+  req->data = luv_setup_req(L, ctx, ref);
   FS_CALL(mkdtemp, req, tpl);
 }
 
 static int luv_fs_rmdir(lua_State* L) {
+  luv_ctx_t* ctx = luv_context(L);
   const char* path = luaL_checkstring(L, 1);
   int ref = luv_check_continuation(L, 2);
   uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, sizeof(*req));
-  req->data = luv_setup_req(L, ref);
+  req->data = luv_setup_req(L, ctx, ref);
   FS_CALL(rmdir, req, path);
 }
 
 static int luv_fs_scandir(lua_State* L) {
+  luv_ctx_t* ctx = luv_context(L);
   const char* path = luaL_checkstring(L, 1);
   int flags = 0; // TODO: find out what these flags are.
   int ref = luv_check_continuation(L, 2);
   uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, sizeof(*req));
-  req->data = luv_setup_req(L, ref);
+  req->data = luv_setup_req(L, ctx, ref);
   FS_CALL(scandir, req, path, flags);
 }
 
@@ -409,7 +525,6 @@ static int luv_fs_scandir_next(lua_State* L) {
   uv_fs_t* req = luv_check_fs(L, 1);
   uv_dirent_t ent;
   int ret = uv_fs_scandir_next(req, &ent);
-  const char* type;
   if (ret == UV_EOF) {
     luv_cleanup_req(L, (luv_req_t*)req->data);
     req->data = NULL;
@@ -417,148 +532,150 @@ static int luv_fs_scandir_next(lua_State* L) {
     return 0;
   }
   if (ret < 0) return luv_error(L, ret);
-  lua_pushstring(L, ent.name);
-  switch (ent.type) {
-    case UV_DIRENT_UNKNOWN: return 1;
-    case UV_DIRENT_FILE:    type = "file"; break;
-    case UV_DIRENT_DIR:     type = "directory"; break;
-    case UV_DIRENT_LINK:    type = "link"; break;
-    case UV_DIRENT_FIFO:    type = "fifo"; break;
-    case UV_DIRENT_SOCKET:  type = "socket"; break;
-    case UV_DIRENT_CHAR:    type = "char"; break;
-    case UV_DIRENT_BLOCK:   type = "block"; break;
-    default:                type = "unknown"; break;
-  }
-  lua_pushstring(L, type);
-  return 2;
+  return luv_push_dirent(L, &ent, 0);
 }
 
 static int luv_fs_stat(lua_State* L) {
+  luv_ctx_t* ctx = luv_context(L);
   const char* path = luaL_checkstring(L, 1);
   int ref = luv_check_continuation(L, 2);
   uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, sizeof(*req));
-  req->data = luv_setup_req(L, ref);
+  req->data = luv_setup_req(L, ctx, ref);
   FS_CALL(stat, req, path);
 }
 
 static int luv_fs_fstat(lua_State* L) {
+  luv_ctx_t* ctx = luv_context(L);
   uv_file file = luaL_checkinteger(L, 1);
   int ref = luv_check_continuation(L, 2);
   uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, sizeof(*req));
-  req->data = luv_setup_req(L, ref);
+  req->data = luv_setup_req(L, ctx, ref);
   FS_CALL(fstat, req, file);
 }
 
 static int luv_fs_lstat(lua_State* L) {
+  luv_ctx_t* ctx = luv_context(L);
   const char* path = luaL_checkstring(L, 1);
   int ref = luv_check_continuation(L, 2);
   uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, sizeof(*req));
-  req->data = luv_setup_req(L, ref);
+  req->data = luv_setup_req(L, ctx, ref);
   FS_CALL(lstat, req, path);
 }
 
 static int luv_fs_rename(lua_State* L) {
+  luv_ctx_t* ctx = luv_context(L);
   const char* path = luaL_checkstring(L, 1);
   const char* new_path = luaL_checkstring(L, 2);
   int ref = luv_check_continuation(L, 3);
   uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, sizeof(*req));
-  req->data = luv_setup_req(L, ref);
+  req->data = luv_setup_req(L, ctx, ref);
   FS_CALL(rename, req, path, new_path);
 }
 
 static int luv_fs_fsync(lua_State* L) {
+  luv_ctx_t* ctx = luv_context(L);
   uv_file file = luaL_checkinteger(L, 1);
   int ref = luv_check_continuation(L, 2);
   uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, sizeof(*req));
-  req->data = luv_setup_req(L, ref);
+  req->data = luv_setup_req(L, ctx, ref);
   FS_CALL(fsync, req, file);
 }
 
 static int luv_fs_fdatasync(lua_State* L) {
+  luv_ctx_t* ctx = luv_context(L);
   uv_file file = luaL_checkinteger(L, 1);
   int ref = luv_check_continuation(L, 2);
   uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, sizeof(*req));
-  req->data = luv_setup_req(L, ref);
+  req->data = luv_setup_req(L, ctx, ref);
   FS_CALL(fdatasync, req, file);
 }
 
 static int luv_fs_ftruncate(lua_State* L) {
+  luv_ctx_t* ctx = luv_context(L);
   uv_file file = luaL_checkinteger(L, 1);
   int64_t offset = luaL_checkinteger(L, 2);
   int ref = luv_check_continuation(L, 3);
   uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, sizeof(*req));
-  req->data = luv_setup_req(L, ref);
+  req->data = luv_setup_req(L, ctx, ref);
   FS_CALL(ftruncate, req, file, offset);
 }
 
 static int luv_fs_sendfile(lua_State* L) {
+  luv_ctx_t* ctx = luv_context(L);
   uv_file out_fd = luaL_checkinteger(L, 1);
   uv_file in_fd = luaL_checkinteger(L, 2);
   int64_t in_offset = luaL_checkinteger(L, 3);
   size_t length = luaL_checkinteger(L, 4);
   int ref = luv_check_continuation(L, 5);
   uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, sizeof(*req));
-  req->data = luv_setup_req(L, ref);
+  req->data = luv_setup_req(L, ctx, ref);
   FS_CALL(sendfile, req, out_fd, in_fd, in_offset, length);
 }
 
 static int luv_fs_access(lua_State* L) {
+  luv_ctx_t* ctx = luv_context(L);
   const char* path = luaL_checkstring(L, 1);
   int amode = luv_check_amode(L, 2);
   int ref = luv_check_continuation(L, 3);
   uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, sizeof(*req));
-  req->data = luv_setup_req(L, ref);
+  req->data = luv_setup_req(L, ctx, ref);
   FS_CALL(access, req, path, amode);
 }
 
 static int luv_fs_chmod(lua_State* L) {
+  luv_ctx_t* ctx = luv_context(L);
   const char* path = luaL_checkstring(L, 1);
   int mode = luaL_checkinteger(L, 2);
   int ref = luv_check_continuation(L, 3);
   uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, sizeof(*req));
-  req->data = luv_setup_req(L, ref);
+  req->data = luv_setup_req(L, ctx, ref);
   FS_CALL(chmod, req, path, mode);
 }
 
 static int luv_fs_fchmod(lua_State* L) {
+  luv_ctx_t* ctx = luv_context(L);
   uv_file file = luaL_checkinteger(L, 1);
   int mode = luaL_checkinteger(L, 2);
   int ref = luv_check_continuation(L, 3);
   uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, sizeof(*req));
-  req->data = luv_setup_req(L, ref);
+  req->data = luv_setup_req(L, ctx, ref);
   FS_CALL(fchmod, req, file, mode);
 }
 
 static int luv_fs_utime(lua_State* L) {
+  luv_ctx_t* ctx = luv_context(L);
   const char* path = luaL_checkstring(L, 1);
   double atime = luaL_checknumber(L, 2);
   double mtime = luaL_checknumber(L, 3);
   int ref = luv_check_continuation(L, 4);
   uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, sizeof(*req));
-  req->data = luv_setup_req(L, ref);
+  req->data = luv_setup_req(L, ctx, ref);
   FS_CALL(utime, req, path, atime, mtime);
 }
 
 static int luv_fs_futime(lua_State* L) {
+  luv_ctx_t* ctx = luv_context(L);
   uv_file file = luaL_checkinteger(L, 1);
   double atime = luaL_checknumber(L, 2);
   double mtime = luaL_checknumber(L, 3);
   int ref = luv_check_continuation(L, 4);
   uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, sizeof(*req));
-  req->data = luv_setup_req(L, ref);
+  req->data = luv_setup_req(L, ctx, ref);
   FS_CALL(futime, req, file, atime, mtime);
 }
 
 static int luv_fs_link(lua_State* L) {
+  luv_ctx_t* ctx = luv_context(L);
   const char* path = luaL_checkstring(L, 1);
   const char* new_path = luaL_checkstring(L, 2);
   int ref = luv_check_continuation(L, 3);
   uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, sizeof(*req));
-  req->data = luv_setup_req(L, ref);
+  req->data = luv_setup_req(L, ctx, ref);
   FS_CALL(link, req, path, new_path);
 }
 
 static int luv_fs_symlink(lua_State* L) {
+  luv_ctx_t* ctx = luv_context(L);
   const char* path = luaL_checkstring(L, 1);
   const char* new_path = luaL_checkstring(L, 2);
   int flags = 0, ref;
@@ -573,49 +690,67 @@ static int luv_fs_symlink(lua_State* L) {
   }
   ref = luv_check_continuation(L, 4);
   req = (uv_fs_t*)lua_newuserdata(L, sizeof(*req));
-  req->data = luv_setup_req(L, ref);
+  req->data = luv_setup_req(L, ctx, ref);
 
   FS_CALL(symlink, req, path, new_path, flags);
 }
 
 static int luv_fs_readlink(lua_State* L) {
+  luv_ctx_t* ctx = luv_context(L);
   const char* path = luaL_checkstring(L, 1);
   int ref = luv_check_continuation(L, 2);
   uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, sizeof(*req));
-  req->data = luv_setup_req(L, ref);
+  req->data = luv_setup_req(L, ctx, ref);
   FS_CALL(readlink, req, path);
 }
 
 static int luv_fs_realpath(lua_State* L) {
+  luv_ctx_t* ctx = luv_context(L);
   const char* path = luaL_checkstring(L, 1);
   int ref = luv_check_continuation(L, 2);
   uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, sizeof(*req));
-  req->data = luv_setup_req(L, ref);
+  req->data = luv_setup_req(L, ctx, ref);
   FS_CALL(realpath, req, path);
 }
 
 static int luv_fs_chown(lua_State* L) {
+  luv_ctx_t* ctx = luv_context(L);
   const char* path = luaL_checkstring(L, 1);
   uv_uid_t uid = luaL_checkinteger(L, 2);
   uv_uid_t gid = luaL_checkinteger(L, 3);
   int ref = luv_check_continuation(L, 4);
   uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, sizeof(*req));
-  req->data = luv_setup_req(L, ref);
+  req->data = luv_setup_req(L, ctx, ref);
   FS_CALL(chown, req, path, uid, gid);
 }
 
 static int luv_fs_fchown(lua_State* L) {
+  luv_ctx_t* ctx = luv_context(L);
   uv_file file = luaL_checkinteger(L, 1);
   uv_uid_t uid = luaL_checkinteger(L, 2);
   uv_uid_t gid = luaL_checkinteger(L, 3);
   int ref = luv_check_continuation(L, 4);
   uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, sizeof(*req));
-  req->data = luv_setup_req(L, ref);
+  req->data = luv_setup_req(L, ctx, ref);
   FS_CALL(fchown, req, file, uid, gid);
 }
 
+#if LUV_UV_VERSION_GEQ(1, 21, 0)
+static int luv_fs_lchown(lua_State* L) {
+  luv_ctx_t* ctx = luv_context(L);
+  const char* path = luaL_checkstring(L, 1);
+  uv_uid_t uid = luaL_checkinteger(L, 2);
+  uv_uid_t gid = luaL_checkinteger(L, 3);
+  int ref = luv_check_continuation(L, 4);
+  uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, sizeof(*req));
+  req->data = luv_setup_req(L, ctx, ref);
+  FS_CALL(lchown, req, path, uid, gid);
+}
+#endif
+
 #if LUV_UV_VERSION_GEQ(1, 14, 0)
 static int luv_fs_copyfile(lua_State*L) {
+  luv_ctx_t* ctx = luv_context(L);
   const char* path = luaL_checkstring(L, 1);
   const char* new_path = luaL_checkstring(L, 2);
   int flags = 0, ref;
@@ -638,7 +773,89 @@ static int luv_fs_copyfile(lua_State*L) {
   }
   ref = luv_check_continuation(L, 4);
   req = (uv_fs_t*)lua_newuserdata(L, sizeof(*req));
-  req->data = luv_setup_req(L, ref);
+  req->data = luv_setup_req(L, ctx, ref);
   FS_CALL(copyfile, req, path, new_path, flags);
 }
 #endif
+
+#if LUV_UV_VERSION_GEQ(1, 28, 0)
+static uv_dir_t* luv_check_dir(lua_State* L, int idx) {
+  uv_dir_t* dir = *(uv_dir_t**)luaL_checkudata(L, idx, "uv_dir");
+  return dir;
+}
+
+static int luv_fs_opendir(lua_State* L) {
+  luv_ctx_t* ctx = luv_context(L);
+  const char* path = luaL_checkstring(L, 1);
+  int ref = luv_check_continuation(L, 2);
+  size_t nentries = luaL_optinteger(L, 3, 1);
+  uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, sizeof(*req));
+  req->data = luv_setup_req(L, ctx, ref);
+
+  //make data_ref to nentries
+  lua_pushinteger(L, nentries);
+  ((luv_req_t*)req->data)->data_ref = luaL_ref(L, LUA_REGISTRYINDEX);
+
+  FS_CALL(opendir, req, path);
+}
+
+static int luv_fs_readdir(lua_State* L) {
+  luv_ctx_t* ctx = luv_context(L);
+  uv_fs_t *req;
+  uv_dir_t* dir = luv_check_dir(L, 1);
+  int ref = luv_check_continuation(L, 2);
+
+  req = (uv_fs_t*)lua_newuserdata(L, sizeof(*req));
+  req->data = luv_setup_req(L, ctx, ref);
+  FS_CALL(readdir, req, dir);
+}
+
+static int luv_fs_closedir(lua_State* L) {
+  luv_ctx_t* ctx = luv_context(L);
+  uv_dir_t* dir = luv_check_dir(L, 1);
+  int ref = luv_check_continuation(L, 2);
+  uv_fs_t *req = (uv_fs_t*)lua_newuserdata(L, sizeof(*req));
+  req->data = luv_setup_req(L, ctx, ref);
+  lua_pushfstring(L, "uv_dir:%p", dir);
+  lua_pushnil(L);
+  lua_rawset(L, LUA_REGISTRYINDEX);
+  FS_CALL(closedir, req, dir);
+}
+
+static int luv_fs_dir_tostring(lua_State* L) {
+  uv_dir_t* dir = luv_check_dir(L, 1);
+  lua_pushfstring(L, "uv_dir_t: %p", dir);
+  return 1;
+}
+
+static int luv_fs_dir_gc(lua_State* L) {
+  uv_dir_t* dir = luv_check_dir(L, 1);
+  lua_pushfstring(L, "uv_dir:%p", dir);
+  lua_rawget(L, LUA_REGISTRYINDEX);
+  if (!lua_isnil(L, -1)) {
+    uv_fs_t req;
+    luv_ctx_t* ctx = luv_context(L);
+
+    uv_fs_closedir(ctx->loop, &req, dir, NULL);
+    uv_fs_req_cleanup(&req);
+    lua_pushfstring(L, "uv_dir:%p", dir);
+    lua_pushnil(L);
+    lua_rawset(L, LUA_REGISTRYINDEX);
+  }
+  lua_pop(L, 1);
+
+  return 0;
+}
+#endif
+
+#if LUV_UV_VERSION_GEQ(1, 31, 0)
+static int luv_fs_statfs(lua_State* L) {
+  luv_ctx_t* ctx = luv_context(L);
+  const char* path = luaL_checkstring(L, 1);
+  int ref = luv_check_continuation(L, 2);
+  uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, sizeof(*req));
+  req->data = luv_setup_req(L, ctx, ref);
+  FS_CALL(statfs, req, path);
+}
+#endif
+
