@@ -15,7 +15,13 @@
  *
  */
 
+#include <lua.h>
+#if (LUA_VERSION_NUM != 503)
+#include "c-api/compat-5.3.h"
+#endif
+#define LUV_SOURCE
 #include "luv.h"
+
 #include "util.c"
 #include "lhandle.c"
 #include "lreq.c"
@@ -129,6 +135,9 @@ static const luaL_Reg luv_functions[] = {
   {"is_readable", luv_is_readable},
   {"is_writable", luv_is_writable},
   {"stream_set_blocking", luv_stream_set_blocking},
+#if LUV_UV_VERSION_GEQ(1, 19, 0)
+  {"stream_get_write_queue_size", luv_stream_get_write_queue_size},
+#endif
 
   // tcp.c
   {"new_tcp", luv_new_tcp},
@@ -141,6 +150,9 @@ static const luaL_Reg luv_functions[] = {
   {"tcp_getsockname", luv_tcp_getsockname},
   {"tcp_connect", luv_tcp_connect},
   {"tcp_write_queue_size", luv_write_queue_size},
+#if LUV_UV_VERSION_GEQ(1, 32, 0)
+  {"tcp_close_reset", luv_tcp_close_reset},
+#endif
 
   // pipe.c
   {"new_pipe", luv_new_pipe},
@@ -176,6 +188,10 @@ static const luaL_Reg luv_functions[] = {
   {"udp_try_send", luv_udp_try_send},
   {"udp_recv_start", luv_udp_recv_start},
   {"udp_recv_stop", luv_udp_recv_stop},
+#if LUV_UV_VERSION_GEQ(1, 27, 0)
+  {"udp_connect", luv_udp_connect},
+  {"udp_getpeername", luv_udp_getpeername},
+#endif
 
   // fs_event.c
   {"new_fs_event", luv_new_fs_event},
@@ -219,9 +235,20 @@ static const luaL_Reg luv_functions[] = {
   {"fs_realpath", luv_fs_realpath},
   {"fs_chown", luv_fs_chown},
   {"fs_fchown", luv_fs_fchown},
+#if LUV_UV_VERSION_GEQ(1, 21, 0)
+  {"fs_lchown", luv_fs_lchown},
+#endif
 #if LUV_UV_VERSION_GEQ(1, 14, 0)
   {"fs_copyfile", luv_fs_copyfile },
 #endif
+#if LUV_UV_VERSION_GEQ(1, 28, 0)
+  {"fs_opendir", luv_fs_opendir},
+  {"fs_readdir", luv_fs_readdir},
+  {"fs_closedir", luv_fs_closedir},
+#endif
+#if LUV_UV_VERSION_GEQ(1, 31, 0)
+  {"fs_statfs", luv_fs_statfs},
+#endif
 
   // dns.c
   {"getaddrinfo", luv_getaddrinfo},
@@ -236,6 +263,9 @@ static const luaL_Reg luv_functions[] = {
   {"cwd", luv_cwd},
   {"exepath", luv_exepath},
   {"get_process_title", luv_get_process_title},
+#if LUV_UV_VERSION_GEQ(1, 29, 0)
+  {"get_constrained_memory", luv_get_constrained_memory},
+#endif
   {"get_total_memory", luv_get_total_memory},
   {"get_free_memory", luv_get_free_memory},
   {"getpid", luv_getpid},
@@ -271,7 +301,20 @@ static const luaL_Reg luv_functions[] = {
   {"os_getppid", luv_os_getppid },
 #endif
 #if LUV_UV_VERSION_GEQ(1, 18, 0)
-  {"os_getpid", luv_os_getpid },
+  {"os_getpid", luv_os_getpid},
+#endif
+#if LUV_UV_VERSION_GEQ(1, 23, 0)
+  {"os_getpriority", luv_os_getpriority},
+  {"os_setpriority", luv_os_setpriority},
+#endif
+#if LUV_UV_VERSION_GEQ(1, 25, 0)
+  {"os_uname", luv_os_uname},
+#endif
+#if LUV_UV_VERSION_GEQ(1, 28, 0)
+  {"gettimeofday", luv_gettimeofday},
+#endif
+#if LUV_UV_VERSION_GEQ(1, 31, 0)
+  {"os_environ", luv_os_environ},
 #endif
 
   // thread.c
@@ -350,6 +393,9 @@ static const luaL_Reg luv_stream_methods[] = {
   {"is_readable", luv_is_readable},
   {"is_writable", luv_is_writable},
   {"set_blocking", luv_stream_set_blocking},
+#if LUV_UV_VERSION_GEQ(1, 19, 0)
+  {"get_write_queue_size", luv_stream_get_write_queue_size},
+#endif
   {NULL, NULL}
 };
 
@@ -392,6 +438,9 @@ static const luaL_Reg luv_tcp_methods[] = {
   {"getsockname", luv_tcp_getsockname},
   {"connect", luv_tcp_connect},
   {"write_queue_size", luv_write_queue_size},
+#if LUV_UV_VERSION_GEQ(1, 32, 0)
+  {"close_reset", luv_tcp_close_reset},
+#endif
   {NULL, NULL}
 };
 
@@ -426,6 +475,10 @@ static const luaL_Reg luv_udp_methods[] = {
   {"try_send", luv_udp_try_send},
   {"recv_start", luv_udp_recv_start},
   {"recv_stop", luv_udp_recv_stop},
+#if LUV_UV_VERSION_GEQ(1, 27, 0)
+  {"connect", luv_udp_connect},
+  {"getpeername", luv_udp_getpeername},
+#endif
   {NULL, NULL}
 };
 
@@ -435,6 +488,25 @@ static const luaL_Reg luv_signal_methods[] = {
   {NULL, NULL}
 };
 
+#if LUV_UV_VERSION_GEQ(1, 28, 0)
+static const luaL_Reg luv_dir_methods[] = {
+  {"readdir", luv_fs_readdir},
+  {"closedir", luv_fs_closedir},
+  {NULL, NULL}
+};
+
+static void luv_dir_init(lua_State* L) {
+  luaL_newmetatable(L, "uv_dir");
+  lua_pushcfunction(L, luv_fs_dir_tostring);
+  lua_setfield(L, -2, "__tostring");
+  lua_pushcfunction(L, luv_fs_dir_gc);
+  lua_setfield(L, -2, "__gc");
+  luaL_newlib(L, luv_dir_methods);
+  lua_setfield(L, -2, "__index");
+  lua_pop(L, 1);
+}
+#endif
+
 static void luv_handle_init(lua_State* L) {
 
   lua_newtable(L);
@@ -480,18 +552,99 @@ static void luv_handle_init(lua_State* L) {
   lua_setfield(L, LUA_REGISTRYINDEX, "uv_stream");
 }
 
-LUALIB_API lua_State* luv_state(uv_loop_t* loop) {
-  return (lua_State*)loop->data;
+// Call lua function, will pop nargs values from top of vm stack and push some
+// values according to nresults. When error occurs, it will print error message
+// to stderr, and memory allocation error will cause exit.
+LUALIB_API int luv_cfpcall(lua_State* L, int nargs, int nresult, int flags) {
+  int ret, top, errfunc;
+
+  // Get the traceback function in case of error
+  if ((flags & (LUVF_CALLBACK_NOTRACEBACK|LUVF_CALLBACK_NOERRMSG) ) == 0)
+  {
+    lua_pushcfunction(L, luv_traceback);
+    errfunc = lua_gettop(L);
+    // And insert it before the function and args
+    lua_insert(L, -2 - nargs);
+    errfunc -= (nargs+1);
+  }else
+    errfunc = 0;
+  top  = lua_gettop(L);
+
+  ret = lua_pcall(L, nargs, nresult, errfunc);
+  switch (ret) {
+  case LUA_OK:
+    break;
+  case LUA_ERRMEM:
+    if ((flags & LUVF_CALLBACK_NOERRMSG) == 0)
+      fprintf(stderr, "System Error: %s\n", lua_tostring(L, -1));
+    if ((flags & LUVF_CALLBACK_NOEXIT) == 0)
+      exit(-1);
+    lua_pop(L, 1);
+    ret = -ret;
+    break;
+  case LUA_ERRRUN:
+  case LUA_ERRSYNTAX:
+  case LUA_ERRERR:
+  default:
+    if ((flags & LUVF_CALLBACK_NOERRMSG) == 0)
+      fprintf(stderr, "Uncaught Error: %s\n", lua_tostring(L, -1));
+    lua_pop(L, 1);
+    ret = -ret;
+    break;
+  }
+  if ((flags & (LUVF_CALLBACK_NOTRACEBACK|LUVF_CALLBACK_NOERRMSG) ) == 0)
+  {
+    lua_remove(L, errfunc);
+  }
+  if (ret == LUA_OK) {
+    if(nresult == LUA_MULTRET)
+      nresult = lua_gettop(L) - top + nargs + 1;
+    return nresult;
+  }
+  return ret;
 }
 
-// TODO: find out if storing this somehow in an upvalue is faster
-LUALIB_API uv_loop_t* luv_loop(lua_State* L) {
-  uv_loop_t* loop;
-  lua_pushstring(L, "uv_loop");
+// TODO: see if we can avoid using a string key for this to increase performance
+static const char* luv_ctx_key = "luv_context";
+
+// Please look at luv_ctx_t in luv.h
+LUALIB_API luv_ctx_t* luv_context(lua_State* L) {
+  luv_ctx_t* ctx;
+  lua_pushstring(L, luv_ctx_key);
   lua_rawget(L, LUA_REGISTRYINDEX);
-  loop = (uv_loop_t*)lua_touserdata(L, -1);
+  if (lua_isnil(L, -1)) {
+    // create it if not exist in registry
+    lua_pushstring(L, luv_ctx_key);
+    ctx = (luv_ctx_t*)lua_newuserdata(L, sizeof(*ctx));
+    memset(ctx, 0, sizeof(*ctx));
+    lua_rawset(L, LUA_REGISTRYINDEX);
+  } else {
+    ctx = (luv_ctx_t*)lua_touserdata(L, -1);
+  }
   lua_pop(L, 1);
-  return loop;
+  return ctx;
+}
+
+LUALIB_API lua_State* luv_state(lua_State* L) {
+  return luv_context(L)->L;
+}
+
+LUALIB_API uv_loop_t* luv_loop(lua_State* L) {
+  return luv_context(L)->loop;
+}
+
+// Set an external loop, before luaopen_luv
+LUALIB_API void luv_set_loop(lua_State* L, uv_loop_t* loop) {
+  luv_ctx_t* ctx = luv_context(L);
+
+  ctx->loop = loop;
+  ctx->L = L;
+}
+
+// Set an external event callback routine, before luaopen_luv
+LUALIB_API void luv_set_callback(lua_State* L, luv_CFpcall pcall) {
+  luv_ctx_t* ctx = luv_context(L);
+  ctx->pcall = pcall;
 }
 
 static void walk_cb(uv_handle_t *handle, void *arg)
@@ -503,7 +656,10 @@ static void walk_cb(uv_handle_t *handle, void *arg)
 }
 
 static int loop_gc(lua_State *L) {
-  uv_loop_t* loop = luv_loop(L);
+  luv_ctx_t *ctx = luv_context(L);
+  uv_loop_t* loop = ctx->loop;
+  if (loop==NULL)
+    return 0;
   // Call uv_close on every active handle
   uv_walk(loop, walk_cb, NULL);
   // Run the event loop until all handles are successfully closed
@@ -513,40 +669,54 @@ static int loop_gc(lua_State *L) {
   return 0;
 }
 
-LUALIB_API int luaopen_luv (lua_State *L) {
+LUALIB_API int luaopen_luv (lua_State* L) {
+  luv_ctx_t* ctx = luv_context(L);
 
-  uv_loop_t* loop;
-  int ret;
+  luaL_newlib(L, luv_functions);
 
-  // Setup the uv_loop meta table for a proper __gc
-  luaL_newmetatable(L, "uv_loop.meta");
-  lua_pushstring(L, "__gc");
-  lua_pushcfunction(L, loop_gc);
-  lua_settable(L, -3);
+  // loop is NULL, luv need to create an inner loop
+  if (ctx->loop==NULL) {
+    int ret;
+    uv_loop_t* loop;
 
-  loop = (uv_loop_t*)lua_newuserdata(L, sizeof(*loop));
-  ret = uv_loop_init(loop);
-  if (ret < 0) {
-    return luaL_error(L, "%s: %s\n", uv_err_name(ret), uv_strerror(ret));
+    // Setup the uv_loop meta table for a proper __gc
+    luaL_newmetatable(L, "uv_loop.meta");
+    lua_pushstring(L, "__gc");
+    lua_pushcfunction(L, loop_gc);
+    lua_settable(L, -3);
+    lua_pop(L, 1);
+
+    lua_pushstring(L, "_loop");
+    loop = (uv_loop_t*)lua_newuserdata(L, sizeof(*loop));
+    // setup the userdata's metatable for __gc
+    luaL_getmetatable(L, "uv_loop.meta");
+    lua_setmetatable(L, -2);
+    // create a ref to loop, avoid __gc early
+    // this puts the loop userdata into the _loop key
+    // in the returned luv table
+    lua_rawset(L, -3);
+
+    ctx->loop = loop;
+    ctx->L = L;
+
+    ret = uv_loop_init(loop);
+    if (ret < 0) {
+      return luaL_error(L, "%s: %s\n", uv_err_name(ret), uv_strerror(ret));
+    }
+  }
+  // pcall is NULL, luv use default callback routine
+  if (ctx->pcall==NULL) {
+    ctx->pcall = luv_cfpcall;
   }
-  // setup the metatable for __gc
-  luaL_getmetatable(L, "uv_loop.meta");
-  lua_setmetatable(L, -2);
-  // Tell the state how to find the loop.
-  lua_pushstring(L, "uv_loop");
-  lua_insert(L, -2);
-  lua_rawset(L, LUA_REGISTRYINDEX);
-  lua_pop(L, 1);
-
-  // Tell the loop how to find the state.
-  loop->data = L;
 
   luv_req_init(L);
   luv_handle_init(L);
+#if LUV_UV_VERSION_GEQ(1, 28, 0)
+  luv_dir_init(L);
+#endif
   luv_thread_init(L);
   luv_work_init(L);
 
-  luaL_newlib(L, luv_functions);
   luv_constants(L);
   lua_setfield(L, -2, "constants");
   return 1;
